import re


inside_parenthesis_pattern = re.compile(r"\(.+\)")


def inside_parenthesis(s: str) -> str:
    """Extract a part of the string s between parenthesis.

    Used to extract hint, clarifications or alternate translations within
    dictionary entry.

    Parameters
    ----------
    s : str
        Text to be searched in.

    Returns
    ----------
    t : str
        Text inside parenthesis.
    """
    match = inside_parenthesis_pattern.search(s)
    if match is None:
        return ""
    return match.group()[1:-1]



import os
from pathlib import Path
from typing import Generator, Optional

from prompt_toolkit.document import Document
from prompt_toolkit.completion import Completion, CompleteEvent
from InquirerPy import inquirer
from InquirerPy.prompts.filepath import FilePathCompleter
from InquirerPy.validator import PathValidator


class ExtensionFilePathCompleter(FilePathCompleter):
    """A class used to generate completions for path of a file with an extension.

    Extends `FilePathCompleter` provided by `InquirerPy` library. Overloads the
    `get_completions` method to generate completions only for files with
    specified extension, given the extension is provided.

    Attributes
    ----------
    only_directories : bool
        True, if completions should contain only directories, False by default
    only_files : bool
        True, if completions should contain only directories, False by default.
    extension : Optional[str]
        If given, all files with extensions differing from specified will be
        filtered from completions. By default (None) no filtering is performed.
    """

    def __init__(
        self,
        only_directories: bool = False,
        only_files: bool = False,
        extension: Optional[str] = None,
    ):
        super().__init__(only_directories, only_files)
        self.extension = extension

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Generator[Completion, None, None]:
        "Generator yielding possible path completions."
        all_valid_path_completions = list(
            super().get_completions(document, complete_event)
        )
        for completion in all_valid_path_completions:
            filename = completion.text
            if filename.startswith("."):
                continue
            path = Path(document.current_line, filename)
            if self.extension is not None:
                if path.is_file() and path.suffix != self.extension:
                    continue
            yield completion


def prompt_filepath(
    message: str,
    is_file: bool = False,
    extension: Optional[str] = None,
    directory: Optional[str] = None,
) -> str:
    """Prompt a file path from the user.

    Parameters
    ----------
    message : str
        A text to be printed before the prompt.
    is_file : bool
        True, if the path should lead to an existing file.
    extension : Optional[str]
        The extension of the filepath to be prompted.
    directory : Optional[str]
        A string representing a path to a directory, inside of which the path
        should lead to.

    Returns
    ----------
    path : str
        A string representing a path to a file.
    """
    validator = PathValidator(is_file=True) if is_file else None
    completer = ExtensionFilePathCompleter(extension=extension)
    path = Path(directory) if directory is not None else Path.cwd()
    return inquirer.text(
        message=message,
        default=str(path.absolute()) + os.path.sep,
        completer=completer,
        validate=validator,
    ).execute()


if __name__ == "__main__":
    print(prompt_filepath("> ", extension=".txt", is_file=True))



from enum import Enum
from functools import partial
from typing import Type, Callable
import random

from rich import print
import typer
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from practice_turkish.languages import prompt_way_of_translation
from practice_turkish.filepath import prompt_filepath
from practice_turkish.dictionaries import (
    Dictionary,
    DictionaryEntry,
    CSVDictionaryEntry,
    TurkrutDictionaryEntry,
)


class AnswerType(str, Enum):
    """An enum used to represent form of the answers by an user.

    Values
    ----------
    typing
        User types in an answer using a keyboard.
    choice
        User picks an answer from a list of options.
    """

    TYPING = "TYPING"
    CHOICE = "CHOICE"


def prompt_dictionary_type() -> Type[DictionaryEntry]:
    """Prompt a type of a dictionary from the user.

    Request the user to pick a type of a supported dictionary.
    This type later is used in order to parse the content of the file.

    Returns
    ----------
    T : A subclass of DictionaryItem class
    """
    return inquirer.select(
        message="What is the format of the file?",
        choices=[
            Choice(value=TurkrutDictionaryEntry, name="turkrut.ru"),
            Choice(value=CSVDictionaryEntry, name="csv"),
        ],
    ).execute()


def prompt_shuffle() -> bool:
    """Prompt the user whether to shuffle the dictionary or not.

    Returns
    ----------
    x : bool
        True if user chooses to shuffle, False otherwise.
    """
    return inquirer.select(
        message="What order of questions would you prefer?",
        choices=[
            Choice(value=True, name="Shuffled order"),
            Choice(value=False, name="As it is in the file"),
        ],
    ).execute()


def prompt_answer_type() -> AnswerType:
    "Prompt the user to pick the form of their answers."
    return inquirer.select(
        message="How would you prefer to answer?",
        choices=[
            Choice(value=AnswerType.TYPING, name="Type it in"),
            Choice(value=AnswerType.CHOICE, name="Choose from multiple options"),
        ],
    ).execute()


def answer_with_prompt(entry: DictionaryEntry, a2b: bool) -> bool:
    """Prompt an answer from the user by typing it in, and check its correctness.

    Prompt the user to translate a dictionary item by typing a translation
    in console, and check if the translation is correct.

    Parameters
    ----------
    entry : DictionaryEntry
        The dictionary entry to be practiced.
    a2b : bool
        True, if translation should be checked from language a to language b.
        False, otherwise.

    Returns
    ----------
    is_correct : bool
        True if translation is correct, False otherwise.
    """
    answer = entry.prompt_translation(a2b)
    is_correct = entry.check_translation(a2b, answer)
    correct_translation = entry.query_b if a2b else entry.query_a
    if is_correct:
        print("[green]Correct![/green]", end=" ")
    else:
        print("[red]Incorrect![/red]", end=" ")
    print(f'In the file: "[green]{correct_translation}[/green]".')
    return is_correct


def answer_with_choice(
    the_entry: DictionaryEntry,
    dictionary: Dictionary[DictionaryEntry],
    a2b: bool,
    n_choices: int = 4,
) -> bool:
    """Prompt an answer from the user by picking from several options, check its correctness.

    Prompt the user to translate a dictionary item by picking an answer from
    list of `n_choices` options, and check if the user picked the correct one.

    Parameters
    ----------
    the_entry : DictionaryEntry
        The dictionary entry to be practiced.
    dictionary : Dictionary
        The dictionary other options will be sampled from.
    a2b : bool
        True, if translation should be checked from language a to language b.
        False, otherwise.
    n_choices : int
        The number of options to pick from, default is 4.

    Returns
    ----------
    is_correct : bool
        True if translation is correct, False otherwise.
    """
    query = the_entry.query_a if a2b else the_entry.query_b
    incorrect_options = random.sample(dictionary.entries, k=n_choices)
    incorrect_options = [entry for entry in incorrect_options if entry is not the_entry]
    options = [the_entry] + incorrect_options[: n_choices - 1]
    random.shuffle(options)
    choices = [
        Choice(value=i, name=o.query_b if a2b else o.query_a)
        for i, o in enumerate(options)
    ]

    i = inquirer.select(message=f"{query} ⇨ ", choices=choices).execute()

    choice = options[i]
    if choice is the_entry:
        print("[green]Correct![/green]")
        return True
    correct_answer = the_entry.query_b if a2b else the_entry.query_a
    print(f"[red]Incorrect![/red] Correct option was '[green]{correct_answer}[/green]'")
    return False


def prepare_session() -> (
    tuple[Dictionary[DictionaryEntry], Callable[[DictionaryEntry], bool]]
):
    """Prepare translation session.

    1) Prompts dictionary type, path to it and then loads it.
    2) Prompts form of answering, and prepares answer function

    Returns
    ----------
    dictionary: Dictionary
        Loaded dictionary.
    answer_function: Callable[[DictionaryEntry], bool]
        Function taking in a dictionary entry, prompting user to translate it
        and returning boolean value indicating if the given translation is
        correct.
    """
    dictionary_entry_type = prompt_dictionary_type()
    path = prompt_filepath(
        message="Choose file to practice: ",
        is_file=True,
        extension=dictionary_entry_type.extension(),
        directory=dictionary_entry_type.default_directory(),
    )
    dictionary = Dictionary.from_file(path, dictionary_entry_type)
    if prompt_shuffle():
        dictionary.shuffle()

    a2b = prompt_way_of_translation(dictionary.language_a, dictionary.language_b)
    match prompt_answer_type():
        case AnswerType.TYPING:
            answer_function = partial(answer_with_prompt, a2b=a2b)
        case AnswerType.CHOICE:
            answer_function = partial(
                answer_with_choice, a2b=a2b, dictionary=dictionary
            )

    return dictionary, answer_function


def translation(
    config: str = typer.Option(
        "config.ini", "--config", help="Path to your configuration file."
    )
) -> None:
    """Run a translation session based on a dictionary.

    Parameters
    ----------
    config : str
        A string representing path to your configuration file, default is
        'config.ini'.
    """
    dictionary, answer_function = prepare_session()
    mistakes: Dictionary[DictionaryEntry] = Dictionary(
        [], dictionary.language_a, dictionary.language_b
    )
    for entry in dictionary:
        is_correct = answer_function(entry)
        if not is_correct:
            mistakes.insert(entry)

    mistakes.print(title="Your mistakes")
    total = len(dictionary)
    incorrect = len(mistakes)
    correct = total - incorrect
    print(f"Correct:   [green]{correct:3}[/green]/{total}")
    print(f"Incorrect: [red]{incorrect:3}[/red]/{total}")
    if mistakes:
        if inquirer.confirm(
            message="Send your mistakes to telegram", default=True
        ).execute():
            mistakes.send_to_telegram(config)


def main() -> None:
    """If open as a script, run make translate function."""
    typer.run(translation)


if __name__ == "__main__":
    main()



from string import ascii_letters
from typing import Any
from prompt_toolkit import prompt
from practice_turkish.languages.validator import SymbolValidator


class EnglishValidator(SymbolValidator):
    """A class used to validate an input in english."""

    valid_symbols = set(list(ascii_letters + " "))


def prompt_english(
    message: str = "> ", additional_symbols: str = "", **kwargs: Any
) -> str:
    """Prompt an input in English from the user.

    Parameters
    ----------
    message : str
        A text to be printed before the prompt. Defaults is "> ".
    additional_symbols : str
        A string of symbols, which should be considered valid, in addition
        to alphabet symbols and space.

    Returns
    ----------
    s : str
        A string typed in by the user.
    """
    return prompt(
        message,
        validator=EnglishValidator(additional_symbols),
        complete_while_typing=False,
        mouse_support=True,
        **kwargs
    ).strip()


if __name__ == "__main__":
    prompt_english("Test the english prompt: ")



import ctypes

from ctypes.wintypes import BOOL, HWND, HANDLE, HGLOBAL, UINT, LPVOID
from ctypes import c_size_t as SIZE_T

OpenClipboard = ctypes.windll.user32.OpenClipboard
OpenClipboard.argtypes = (HWND,)
OpenClipboard.restype = BOOL
EmptyClipboard = ctypes.windll.user32.EmptyClipboard
EmptyClipboard.restype = BOOL
GetClipboardData = ctypes.windll.user32.GetClipboardData
GetClipboardData.argtypes = (UINT,)
GetClipboardData.restype = HANDLE
SetClipboardData = ctypes.windll.user32.SetClipboardData
SetClipboardData.argtypes = UINT, HANDLE
SetClipboardData.restype = HANDLE
CloseClipboard = ctypes.windll.user32.CloseClipboard
CloseClipboard.restype = BOOL
CF_UNICODETEXT = 13

GlobalAlloc = ctypes.windll.kernel32.GlobalAlloc
GlobalAlloc.argtypes = UINT, SIZE_T
GlobalAlloc.restype = HGLOBAL
GlobalLock = ctypes.windll.kernel32.GlobalLock
GlobalLock.argtypes = (HGLOBAL,)
GlobalLock.restype = LPVOID
GlobalUnlock = ctypes.windll.kernel32.GlobalUnlock
GlobalUnlock.argtypes = (HGLOBAL,)
GlobalSize = ctypes.windll.kernel32.GlobalSize
GlobalSize.argtypes = (HGLOBAL,)
GlobalSize.restype = SIZE_T

GMEM_MOVEABLE = 0x0002
GMEM_ZEROINIT = 0x0040


def copy_to_clipboard(s: str) -> None:
    """Copies a content of a given string to clipboard.

    Parameters
    ----------
    s : str
        The string to copy from.
    """
    data = s.encode("utf-16le")
    OpenClipboard(None)
    EmptyClipboard()
    handle = GlobalAlloc(GMEM_MOVEABLE | GMEM_ZEROINIT, len(data) + 2)
    pcontents = GlobalLock(handle)
    ctypes.memmove(pcontents, data, len(data))
    GlobalUnlock(handle)
    SetClipboardData(CF_UNICODETEXT, handle)
    CloseClipboard()



from abc import ABC, abstractmethod
from typing import Type, TypeVar, Optional, Iterator, Generic
from dataclasses import dataclass
from random import shuffle

from rich import print
from rich.table import Table

from practice_turkish.languages import Language, PrompterInTheLanguage
from practice_turkish.dictionaries.telegram import (
    APIConfiguration,
    send_to_telegram,
    TelegramError,
)

DE = TypeVar("DE", bound="DictionaryEntry")
D = TypeVar("D", bound="Dictionary")


class DictionaryFormatError(ValueError):
    """Exception raised if dictionary file violates specification."""


class DictionaryEntry(ABC):
    """An ABC used to represent an entry from a dictionary of an arbitrary format.

    Abstract base class for all classes used to represent entries from dictionaries of
    concrete type.

    Prompting an answer and checking if its correct already implemented
    here. Each subclass should implement all the following properties and methods.

    Properties
    ----------
    language_a : Language
        The 1st language (language A below) of the dictionary.
    language_b : Language
        The 2nd language (language B below) of the dictionary.
    query_a : str
        The query to be showed when prompted to translate from language A to
        language B.
    query_b : str
        The query to be showed when prompted to translate from language B to
        language A.
    words_a : set[str]
        Set of options to be considered correct when prompted to translate from
        language B to language A.
    words_b : set[str]
        Set of options to be considered correct when prompted to translate from
        language A to language B.

    Methods
    ----------
    @classmethod
    def extension(cls) -> str
        Returns the file extension used with dictionary of this format.

    @staticmethod
    def default_directory() -> str
        Returns a string representing the default directory dictionary of this
        format are stored in.

    @classmethod
    def read_dictionary_from_file(cls, path: str) -> Dictionary
        Read the dictionary of entries of the type from a file.
    """

    def prompt_translation(self, a2b: bool) -> str:
        """Prompt the translation for the entry from the user by typing the answer in.

        Parameters
        ----------
        a2b : bool
            True, if translation should be prompted from language A to
            B language. False otherwise.

        Returns
        ----------
        translation : str
            The string typed in by the user.
        """
        query = self.query_a if a2b else self.query_b
        prompter = PrompterInTheLanguage(self.language_b if a2b else self.language_a)
        return prompter.prompt(f"{query} ⇨ ", additional_symbols=",-")

    def check_translation(self, a2b: bool, translation: str) -> bool:
        """Check translation.

        Parameters
        ----------
        a2b : bool
            True, if translation is given in the language B language A to
            B language, False otherwise.
        translation : str
            A string with translation, typically typed in by the user.

        Returns
        ----------
        x : bool
            True, if the translation is correct, False otherwise.
        """
        translation = translation.strip()
        if translation == "":
            return False

        target = self.words_b if a2b else self.words_a
        return translation in target

    def __lt__(self, other: DE) -> bool:
        """Necessary to sort"""
        return self.words_a < other.words_b

    @property
    @abstractmethod
    def language_a(self) -> Language:
        """First language of the entry."""
        raise NotImplementedError

    @property
    @abstractmethod
    def language_b(self) -> Language:
        """Second language of the entry."""
        raise NotImplementedError

    @property
    @abstractmethod
    def query_a(self) -> str:
        """Query to show when prompting translation to language A."""
        raise NotImplementedError

    @property
    @abstractmethod
    def query_b(self) -> str:
        """Query to show when prompting translation to language A."""
        raise NotImplementedError

    @property
    @abstractmethod
    def words_a(self) -> set[str]:
        """Values to be considered correct translation to language A."""
        raise NotImplementedError

    @property
    @abstractmethod
    def words_b(self) -> set[str]:
        """Values to be considered correct translation to language B."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def extension() -> str:
        """File extension associated with this type of entries."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def default_directory() -> str:
        """Directory name associated with this type of entries."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def read_dictionary_from_file(
        cls: Type[DE], path: str
    ) -> tuple[list[DE], Language, Language]:
        """Read list entries of this type from a file."""
        raise NotImplementedError


@dataclass
class Dictionary(Generic[DE]):
    """A class used to represent a dictionary.

    A dictionary is a list of homogeneous dictionary entries. Instances of the
    class are used in order to run translation sessions. Contains list of
    DictionaryEntry instances, and delegates iteration and indexation to the
    list.

    Attributes
    ----------
    entries : list[DictionaryEntry]
        List of entries.
    language_a : Language
        Language A of the dictionary.
    language_b : Language
        Language B of the dictionary.

    Methods
    ----------
    @classmethod
    def from_file(cls, path: str, T: Type[DictionaryEntry]) -> Dictionary:
        Reads dictionary form a file assuming the type T.

    def print(self, title: Optional[str] = None) -> None:
        Prints the dictionary to stdout in a from of the table.

    def send_to_telegram(self) -> bool:
        Send the dictionary to a telegram user via the bot.
    """

    entries: list[DE]
    language_a: Language
    language_b: Language

    @classmethod
    def from_file(cls: Type[D], path: str, type: Type[DE]) -> D:
        "Read dictionary form a file assuming the type T."
        return cls(*type.read_dictionary_from_file(path))

    def print(self, title: Optional[str] = None) -> None:
        "Print the dictionary to stdout in a from of the table."
        self.sort()
        table = Table(title=title)
        table.add_column(self.language_a, justify="left")
        table.add_column(self.language_b, justify="right")

        for word in self.entries:
            table.add_row(word.query_a, word.query_b)

        print(table)

    def send_to_telegram(self, path: str = "config.ini") -> bool:
        """Send the dictionary to a telegram user via the bot.

        Parameters:
        ----------
        path : str
            A string representing a path to a configuration file.
        """
        self.sort()
        rows = [f"{item.query_a} — {item.query_b}" for item in self]
        text = "\n".join(rows)
        try:
            config = APIConfiguration.read_ini(path)
            status = send_to_telegram(config.url, config.user_id, config.token, text)
        except TelegramError as exception:
            exception_type = type(exception).__name__
            print(f"[red]{exception_type}[/red]: [yellow]{exception}[/yellow]")
            return False
        return status

    def sort(self) -> None:
        "Sort the dictionary with respect to the language A."
        self.entries.sort(key=lambda item: item.query_a)

    def insert(self, entry: DE) -> None:
        "Insert a new entry."
        self.entries.append(entry)

    def shuffle(self) -> None:
        "Shuffle entries."
        shuffle(self.entries)

    def __iter__(self) -> Iterator[DE]:
        return iter(self.entries)

    def __len__(self) -> int:
        return len(self.entries)

    def __getitem__(self, index: int) -> DE:
        return self.entries[index]



from practice_turkish.dictionaries.dictionary import (
    Dictionary,
    DictionaryEntry,
    DictionaryFormatError,
)
from practice_turkish.dictionaries.csvdictionary import CSVDictionaryEntry
from practice_turkish.dictionaries.turkrutdictionary import TurkrutDictionaryEntry



from string import ascii_letters
from typing import Generator, Any

from prompt_toolkit import prompt
from prompt_toolkit.document import Document
from prompt_toolkit.completion import Completer, Completion, CompleteEvent

from practice_turkish.languages.clipboard import copy_to_clipboard
from practice_turkish.languages.validator import SymbolValidator

non_latin_letters = {
    "a": "â",
    "c": "ç",
    "C": "Ç",
    "g": "ğ",
    "G": "Ğ",
    "i": "ı",
    "I": "İ",
    "o": "ö",
    "O": "Ö",
    "s": "ş",
    "S": "Ş",
    "u": "ü",
    "U": "Ü",
}


class TurkishCompleter(Completer):
    """A class used in order to type nonlatin letters from Turkish alphabet.

    Used by the `prompt` function from `prompt_toolkit` library in order to
    generate completions, which allow to type in all letters from turkish
    alphabet with only en-US layout keyboard. Pressing tab after letters
    'c', 'g', 'i', 'o', 's' and 'u' offers completion with "ç", "ğ", "ı",
    "ö", "ş" and "ü".
    """

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Generator[Completion, None, None]:
        "Generator yielding completions for the symbol right before the cursor"
        active_letter = document.char_before_cursor
        try:
            suggestion = non_latin_letters[active_letter]
        except KeyError:
            return
        yield Completion(suggestion, start_position=-1)


class TurkishValidator(SymbolValidator):
    """A class used to validate an input in turkish."""

    latin_letters = set(ascii_letters) - set("qQxXwW")
    non_latin_letters = set("âçÇğĞıIiİöÖşŞüÜ")
    valid_symbols = latin_letters | non_latin_letters | {" "}


def prompt_turkish(
    message: str = "> ", additional_symbols: str = "", **kwargs: Any
) -> str:
    """Prompt an input in turkish from the user.

    Prompts an input in Turkish from the user.
    Pressing TAB after letters 'c', 'g', 'i', 'o', 's' and 'u' will offer
    similar looking letters from Turkish alphabet as a replacement.

    Parameters
    ----------
    message : str
        A text to be printed before the prompt. Defaults is "> ".
    additional_symbols : str
        A string of symbols, which should be considered valid, in addition
        to alphabet symbols and space.

    Returns
    ----------
    s : str
        A string typed in by the user.
    """
    return prompt(
        message,
        completer=TurkishCompleter(),
        validator=TurkishValidator(additional_symbols),
        complete_while_typing=False,
        mouse_support=True,
        **kwargs
    ).strip()


def main() -> None:
    """If open as a script, run make number function."""
    while True:
        input_ = prompt_turkish("Test the Turkish prompt: ")
        if input_ == "":
            return
        copy_to_clipboard(input_)


if __name__ == "__main__":
    main()



from configparser import ConfigParser, SectionProxy
from dataclasses import dataclass
import json
from typing import ClassVar

import requests


class TelegramError(Exception):
    "Generic error with regard to the bot API."


class ConfigurationError(TelegramError):
    "Raised when an error detected before trying to use it."


class AuthenticationError(TelegramError):
    "Raised when a pair of (user_id, token) is rejected by the bot."


@dataclass
class APIConfiguration:
    """
    A class used to represent a content of a configuration file.

    Attributes
    ----------
    url : str
        The url of the bot API.
        Should be correct in the `config.ini` file at the github repository.
    user_id : int
        The telegram ID number of the user messages are to be sent to.
        Should be an integer.
    token : str
        The token of the user messages are to be sent to.
        Should be a string of 8 symbols: latin letters (any case) and digits.
    """

    url: ClassVar[str] = "https://redirectfunction-d2ooxt72na-lm.a.run.app"
    user_id: int
    token: str

    @classmethod
    def read_ini(cls, path: str = "config.ini") -> "APIConfiguration":
        """Read configuration file from a file.

        Parameters
        ----------
        path : str
            A string representing a path to the configuration file.

        Raises
        ----------
        ConfigurationError
            If an error ocurred while reading the configuration from a file.
        """
        config = ConfigParser()
        if not config.read(path):
            raise ConfigurationError(f"Missing '{path}' configuration file!")

        try:
            API_section = config["BOT API"]
        except KeyError:
            raise ConfigurationError(
                f"Configuration file '{path}' is incorrect: missing 'BOT API' section."
            )

        user_id = _get_and_check_int(API_section, "USER ID", path)
        token = _get_and_check_str(API_section, "TOKEN", path)

        return cls(user_id, token)


def send_to_telegram(
    url: str,
    user_id: int,
    token: str,
    text: str,
) -> bool:
    """
    Sends message to a telegram user via the bot.

    Parameters
    ----------
    url : str
        The url of the bot.
    user_id : int
        The telegram ID number of the user messages are to be sent to.
        Should be an integer.
    token : str
        The token of the user messages are to be sent to.
        Should be a string of 8 symbols: latin letters (any case) and digits.
    text : str
        Text to send.

    Returns
    ----------
    x : bool
        True if no errors are detected, False otherwise.

    Raises
    ----------
    AuthenticationError
        If the pair (`user_id`, `token`) is rejected by the bot.
    """
    response = requests.post(
        url=url,
        data=json.dumps({"user id": user_id, "token": token, "text": text}),
        timeout=10,
    )
    if response.status_code == 403:
        raise AuthenticationError(
            """User token is rejected!
            Please, check 'USER ID' and 'TOKEN' fields in your configuration file."""
        )
    if response.status_code != 200:
        raise requests.HTTPError(response.text)
    return True


def _get_and_check_int(section: SectionProxy, field: str, path: str) -> int:
    try:
        value = section.getint(field)
    except ValueError:
        raise ConfigurationError(
            f"Configuration file '{path}' is incorrect: '{field}' expected to be an integer."
        )
    if value is None:
        raise ConfigurationError(
            f"Configuration file '{path}' is incorrect: missing '{field}' field."
        )
    return value


def _get_and_check_str(section: SectionProxy, field: str, path: str) -> str:
    try:
        value = section[field]
    except KeyError:
        raise ConfigurationError(
            f"Configuration file {path} is incorrect: missing '{field}' field."
        )
    return value



from typing import Optional, Type, TypeVar, Iterable
from dataclasses import dataclass
import csv

from practice_turkish.languages import Language
from practice_turkish.dictionaries import DictionaryEntry, DictionaryFormatError

DE = TypeVar("DE", bound="CSVDictionaryEntry")


def generate_query(words: Iterable[str], hint: Optional[str]) -> str:
    "Generate query with words separated by '/' and a possible hint."
    query = "/".join(words)
    if hint is not None:
        query += f" ({hint})"
    return query


def parse_language(language: str) -> Language:
    """Parse language from a string and create an instance of Language enum.

    Parameters
    ----------
    language : str
        A string representing a language. Should be equal to one of the values
        of Language enum.

    Raises
    ----------
    DictionaryFormatError
        If the string isn't a value of Language enum.

    """
    normalized = language.strip().lower()
    try:
        return Language[normalized]
    except KeyError:
        raise DictionaryFormatError(f"Not supported language {language}.")


def parse_header(header: list[str]) -> tuple[Language, Language]:
    """Parse languages of the dictionary based on the header of the CSV file.

    Parses languages of the dictionary based on the names of the first two
    columns in the dictionary and returns them in a tuple.

    Parameters
    ----------
    header : list[str]
        A list of column names in a CSV file. Should contain exactly 4 strings.

    Returns
    ----------
    language_a : Language
        Language of the 1st column.
    language_b : Language
        Language of the 2nd column.

    Raises
    ----------
    DictionaryFormatError
        If the number of columns is incorrect or languages can't be parsed.
    """
    if len(header) != 4:
        raise DictionaryFormatError("Incorrect number of columns in the CSV file.")
    lang_a_header, lang_b_header, _, _ = header
    return parse_language(lang_a_header), parse_language(lang_b_header)


@dataclass
class CSVDictionaryEntry(DictionaryEntry):
    """A class used to represent entries of custom dictionary form.

    This form uses a CSV file format with 4 columns to store dictionaries.
    A semicolon (";") is used as the separator of the columns.

    First two columns are mandatory for each entry. They represent correct
    translations from one language to another. Multiple alternative
    translations may be present, in which case they are separated
    by slash ("/") within one cell. Each of the values will be considered
    as a correct translation.

    Other two optional columns may contain possible hints or clarifications
    en each language. They will be shown to the user within a query for
    translation inside parenthesis, but won't be considered as a correct
    translation.

    For meaning of the properties and methods, see its parent ABC.
    """

    _words_a: list[str]
    _words_b: list[str]
    _language_a: Language
    _language_b: Language
    _hint_a: Optional[str] = None
    _hint_b: Optional[str] = None

    @property
    def language_a(self) -> Language:
        return self._language_a

    @property
    def language_b(self) -> Language:
        return self._language_b

    @property
    def words_a(self) -> set[str]:
        return set(self._words_a)

    @property
    def words_b(self) -> set[str]:
        return set(self._words_b)

    @property
    def query_a(self) -> str:
        return generate_query(self.words_a, self._hint_a)

    @property
    def query_b(self) -> str:
        return generate_query(self.words_b, self._hint_b)

    @staticmethod
    def extension() -> str:
        return ".csv"

    @staticmethod
    def default_directory() -> str:
        return "CSV"

    @classmethod
    def read_dictionary_from_file(
        cls: Type[DE], path: str
    ) -> tuple[list[DE], Language, Language]:
        with open(path, encoding="utf-8") as f:
            reader = csv.reader(f, delimiter=";")
            header = next(reader)
            language_a, language_b = parse_header(header)

            dictionary: list[DE] = []
            for words_a, words_b, hint_a, hint_b in reader:
                dictionary.append(
                    cls(
                        words_a.split("/"),
                        words_b.split("/"),
                        language_a,
                        language_b,
                        None if not hint_a else hint_a,
                        None if not hint_b else hint_b,
                    )
                )
        return dictionary, language_a, language_b



import re
from typing import Type, Optional, TypeVar
from dataclasses import dataclass

from practice_turkish.languages import Language
from practice_turkish.dictionaries import DictionaryEntry
from practice_turkish.dictionaries.parse import inside_parenthesis

T = TypeVar("T", bound="TurkrutDictionaryEntry")


def extract_words_and_hint(s: str) -> tuple[set[str], str]:
    """Extracts words and hint for one language

    Parameters
    ----------
    s : str
        A part of a turkrut dictionary line to one side of the middle dash.

    Returns
    ----------
    words : list[str]
        Values to be considered correct translation to the language.
    hint: str
        Hint for translation to the language.
    """
    hint = inside_parenthesis(s)
    words_part = s.replace(f"({hint})", "").strip()
    words = set(words_part.split(", "))
    return words, hint


@dataclass
class TurkrutDictionaryEntry(DictionaryEntry):
    """A class used to represent a dictionary entry from turkrut.ru.

    The structure:
    Turkish (hint) - Russian (hint)\n
    Caveats:
    1) The separator could be a dash of various length.
    2) Might include multiple options for russian translation.
    3) Might include additional info inside parenthesis for both languages.
    """

    _turkish: str
    _russian: str
    _turkish_words: set[str]
    _russian_words: set[str]
    _turkish_hint: Optional[str]
    _russian_hint: Optional[str]

    @property
    def language_a(self) -> Language:
        return Language.turkish

    @property
    def language_b(self) -> Language:
        return Language.russian

    @property
    def query_a(self) -> str:
        return self._turkish

    @property
    def query_b(self) -> str:
        return self._russian

    @property
    def words_a(self) -> set[str]:
        return self._turkish_words

    @property
    def words_b(self) -> set[str]:
        return self._russian_words

    @staticmethod
    def extension() -> str:
        return ".txt"

    @staticmethod
    def default_directory() -> str:
        return "turkrut"

    @classmethod
    def from_line(cls: Type[T], line: str) -> T:
        """Factory class method parsing a line of a turkrut dictionary file.

        Parameters
        ----------
        line : str
            A line of a turkrut dictionary file.

        Returns
        ----------
        item : TurkrutDictionaryItem
            An entry representing the line
        """
        tk, ru = re.split("-|—|–", line)
        tk, ru = tk.strip(), ru.strip()
        tk_words, tk_hint = extract_words_and_hint(tk)
        ru_words, ru_hint = extract_words_and_hint(ru)
        return cls(tk, ru, tk_words, ru_words, tk_hint, ru_hint)

    @classmethod
    def read_dictionary_from_file(
        cls: Type[T], path: str
    ) -> tuple[list[T], Language, Language]:
        with open(path, encoding="utf-8") as f:
            return (
                [cls.from_line(line) for line in f],
                Language.turkish,
                Language.russian,
            )



from enum import Enum
from string import Template
import os
from typing import Optional, TypeAlias
import csv

import typer
from rich import print
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from practice_turkish.languages import Language, PrompterInTheLanguage, prompt_language
from practice_turkish.dictionaries import Dictionary, CSVDictionaryEntry
from practice_turkish.filepath import prompt_filepath
from practice_turkish.dictionaries.parse import inside_parenthesis


CSVDict: TypeAlias = Dictionary[CSVDictionaryEntry]


class WritingMode(str, Enum):
    """Enum used to represent writing mode."""

    EXTEND = "EXTEND"
    CREATE = "CREATE"
    EXIT = "EXIT"
    OVERWRITE = "OVERWRITE"


prompt_text = Template(
    """Type in the word or words in [green]$language[/green].
Text inside parenthesis would be considered as a hint.
Other values separated by [yellow]"/"[/yellow] would be considered as valid answers.
[red]Empty string[/red] to exit immediately."""
)


def prompt_if_file_exists() -> WritingMode:
    """Prompt the user to choose mode of writing to an existing file.

    Invoked if the user typed in a path leading to an existing file.
    Prompts the user to pick the mode to continue with.
    1) Extend: assuming the path leads to an existing correct CSV dictionary,
    all new entries are to be inserted in the dictionary and saved to the same
    file.
    2) Exit: aborts the program immediately. Safe option, existing file is
    untouched.
    3) Overwrite: ignores the existing file and overwrites its content with
    entries typed in during following session.

    Returns
    ----------
    mode : WritingMode
        WritingMode.extend, WritingMode.exit or WritingMode.overwrite.
    """
    return inquirer.select(
        message="Such file already exists. Do you want to",
        choices=[
            Choice(
                value=WritingMode.EXTEND,
                name="Extend. Only possible if the file is a CSV dictionary.",
            ),
            Choice(value=WritingMode.EXIT, name="Abort the program. No data loss."),
            Choice(
                value=WritingMode.OVERWRITE,
                name="Overwrite the existing file. Possible data loss.",
            ),
        ],
    ).execute()


def prepare_session(path: Optional[str]) -> tuple[CSVDict, str]:
    """Prepares completion session for creating a CSV dictionary.

    Prompts path if not given. Then depending on existence of dictionary there and
    users choice either creates new file, extends existing dictionary, rewrites it,
    or aborts the process.

    Parameters
    ----------
    path : Optional[str]
        A string representing path new dictionary to write to. Will be prompted
        from user, if not given.

    Returns
    ----------
    dictionary : Dictionary[CSVDictionaryEntry]
        Loaded existing dictionary or a new empty dictionary.
    path : str
        A string representing path new dictionary to write to.
    """
    if path is None:
        path = prompt_filepath(
            "Type in destination path: ",
            extension=".csv",
            is_file=False,
            directory=CSVDictionaryEntry.default_directory(),
        )

    mode = WritingMode.CREATE
    if os.path.isfile(path):
        mode = prompt_if_file_exists()
        if mode == WritingMode.EXIT:
            raise SystemExit("Aborting the program")

    match mode:
        case WritingMode.CREATE | WritingMode.OVERWRITE:
            la = prompt_language("Choose first language.")
            lb = prompt_language("Choose second language.")
            dictionary: CSVDict = Dictionary([], la, lb)
        case WritingMode.EXTEND:
            dictionary = Dictionary.from_file(path, CSVDictionaryEntry)
            la, lb = dictionary.language_a, dictionary.language_b
            print(
                f"Detected languages are [green]{la}[/green] and [green]{lb}[/green].\n"
            )
    return dictionary, path


def parse_prompt(s: str) -> tuple[list[str], Optional[str]]:
    """Parse a text typed in by the user.

    Parses a piece of text typed in by the user. Expected template is bellow.
    alternative 1/alternative 2/.../alternative n (hint)
    That is, all alternative translations are expected to be separated by the
    splash symbol ("/"), and hint is expected to be inside of parenthesis at
    the end.

    Parameters
    ----------
    s : str
        Text typed in by the user representing entry in one language.

    Returns
    ----------
    words : list[str]
        List of alternative translation for the entry.
    hint : Optional[str]
        A hint or clarification for the entry if given, None otherwise.
    """
    print(s)
    hint = inside_parenthesis(s)
    words = s.replace(f"({hint})", "").strip().split("/")
    return words, hint if hint else None


def prompt_one_language(language: Language) -> str:
    """Prompt the user to type in a part of an entry corresponding to one language.

    Parameters
    ----------
    language : Language
        The language the prompt should be accepted in.

    Returns
    ----------
    text : str
        Line of text typed in by the user. Guaranteed to be in the language given
        as first parameter.
    """
    print(prompt_text.substitute({"language": language.name}))
    prompter = PrompterInTheLanguage(language)
    return prompter.prompt(additional_symbols=",-/()")


def prompt_dictionary_entry(
    language_a: Language, language_b: Language
) -> Optional[CSVDictionaryEntry]:
    """Prompt the user to type in a dictionary entry.

    Prompts the user to type in a line of text twice. Once for language A,
    and once for language B. If either of the lines typed in by the user
    are empty, None is returned.

    Parameters
    ----------
    language_a : Language
        Language A of the dictionary.

    language_b : Language
        Language B of the dictionary.


    Returns
    ----------
    entry : Optional[CSVDictionaryEntry]
        A complete dictionary entry in given two languages if the user typed in
        translation options for both languages, None otherwise.

    """
    if (words_and_hint := prompt_one_language(language_a)) == "":
        return None
    words_a, hint_a = parse_prompt(words_and_hint)

    if (words_and_hint := prompt_one_language(language_b)) == "":
        return None
    words_b, hint_b = parse_prompt(words_and_hint)

    return CSVDictionaryEntry(words_a, words_b, language_a, language_b, hint_a, hint_b)


def prompt_dictionary(dictionary: CSVDict) -> None:
    """Prompts user to type in all dictionary entries.

    Parameters
    ----------
    dictionary: Dictionary[CSVDictionaryEntry]
        A dictionary to fill in with entries.
    """
    while True:
        entry = prompt_dictionary_entry(dictionary.language_a, dictionary.language_b)
        if entry is None:
            return
        dictionary.insert(entry)


def write_dictionary(dictionary: CSVDict, path: str) -> None:
    """Write dictionary to a CSV file.

    Parameters
    ----------
    dictionary : Dictionary[CSVDictionaryEntry]
        Dictionary to write.
    path : str
        A string representing filepath to write to.
    """
    la, lb = dictionary.language_a, dictionary.language_b
    with open(path, mode="w", encoding="utf-8", newline="") as f:
        field_names = [la.name, lb.name, f"{la} hint", f"{lb} hint"]
        writer = csv.DictWriter(f, delimiter=";", fieldnames=field_names)
        writer.writeheader()
        for item in sorted(dictionary):
            writer.writerow(
                {
                    la.name: "/".join(item.words_a),
                    lb.name: "/".join(item.words_b),
                    f"{la} hint": item._hint_a,
                    f"{lb} hint": item._hint_b,
                }
            )


def make_dictionary(
    path: Optional[str] = typer.Argument(
        None, help="Destination path of the dictionary"
    )
) -> None:
    """Make a CSV dictionary by prompting the user to type in each entry.

    Creates, extends or overwrites an existing CSV dictionary prompting
    the user to type in each entry sequentially. An empty entry indicates
    the end of the dictionary.

    Parameters
    ----------
    path : Optional[str]
        A string representing the filepath, the dictionary is to be written
        to. If None, the filepath is prompted from the user during the
        session.
    """
    dictionary, path = prepare_session(path)
    prompt_dictionary(dictionary)
    dictionary.print("New dictionary")
    write_dictionary(dictionary, path)


def main() -> None:
    """If open as a script, run make dictionary function."""
    typer.run(make_dictionary)


if __name__ == "__main__":
    main()



from abc import ABC
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.document import Document


class SymbolValidator(Validator, ABC):
    """A class used to validate an input in a language.

    Used by the `prompt` function from `prompt_toolkit` library to ensure that
    user types in only symbols from alphabet of the language. Extends `Validator`
    class given by the library and overloads the `validate` method. Each subclass
    should overload `valid_symbols` class attribute.
    """

    valid_symbols: set[str] = set()

    def __init__(self, additional_symbols: str = "") -> None:
        super().__init__()
        self.valid_symbols |= set(list(additional_symbols))

    def validate(self, document: Document) -> None:
        """Check if all typed in symbols are permissible.

        Parameters
        ----------
        document : Document
            A current state of prompting session.

        Raises
        ----------
        ValidationError
            If the document contains prohibited symbols.
        """
        for i, s in enumerate(document.text):
            if s not in self.valid_symbols:
                raise ValidationError(
                    message="This input contains symbols out of Russian alphabet.",
                    cursor_position=i,
                )



import random
from enum import Enum
from functools import partial
from typing import Optional, Callable

from rich import print
import typer
from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from practice_turkish.languages import Language, PrompterInTheLanguage


class Difficulty(str, Enum):
    """An enum used to represent difficulty."""

    DIGITS = "DIGITS"
    TENS = "TENS"
    BASIC = "BASIC"
    ADVANCED = "ADVANCED"


TEN = 10
ONE_HUNDRED = 100
ONE_THOUSAND = 1_000
ONE_MILLION = 1_000_000
ONE_BILLION = 1_000_000_000

digits = {
    0: "sıfır",
    1: "bir",
    2: "iki",
    3: "üç",
    4: "dört",
    5: "beş",
    6: "altı",
    7: "yedi",
    8: "sekiz",
    9: "dokuz",
}

tens = {
    0: "",
    10: "on",
    20: "yirmi",
    30: "otuz",
    40: "kırk",
    50: "elli",
    60: "altmış",
    70: "yetmiş",
    80: "seksen",
    90: "doksan",
}

more = {
    ONE_HUNDRED: "yüz",
    ONE_THOUSAND: "bin",
    ONE_MILLION: "milyon",
    ONE_BILLION: "milyar",
}


def spell_small_number(number: int, dismiss_one: bool = False) -> str:
    """Spell a positive integer number lesser than 1000 in turkish.

    Parameters
    ----------
    number : int
        A positive integer number to spell. 0 <= number <= 999.
    dismiss_one : bool
        Whether to put 'bir' in front or note. Necessary, since
        `bin` used instead of `bir bin`.

    Returns
    ----------
    spelling : str
        A line of text with spelling of the number in turkish.
    """
    if number not in range(1000):
        raise ValueError("The number isn't a positive integer number lesser than 999.")

    if dismiss_one and number == 1:
        return ""

    n_hundreds, reminder = divmod(number, ONE_HUNDRED)
    n_tens, n_ones = divmod(reminder, TEN)

    match n_hundreds:
        case 0:
            hundred = ""
        case 1:
            hundred = "yüz"
        case _:
            hundred = digits[n_hundreds] + " yüz"
    ten = tens[n_tens * 10]
    one = digits[n_ones] if n_ones > 0 else ""
    return f"{hundred} {ten} {one}".strip()


def spell_number(number: int) -> str:
    """Spell a positive integer number lesser than 10^12 in turkish.

    Parameters
    ----------
    number : int
        A positive integer number to spell. 0 <= number <= 10^12.

    Returns
    ----------
    spelling : str
        A line of text with spelling of the number in turkish.
    """
    if number not in range(10**12):
        raise ValueError("Too big number")

    if number == 0:
        return digits[0]

    n_billions, reminder = divmod(number, ONE_BILLION)
    n_millions, reminder = divmod(reminder, ONE_MILLION)
    n_thousands, reminder = divmod(reminder, ONE_THOUSAND)

    parts = []
    if n_billions:
        parts.append(f"{spell_small_number(n_billions)} milyar".strip())

    if n_millions:
        parts.append(f"{spell_small_number(n_millions)} milyon".strip())

    if n_thousands:
        parts.append(f"{spell_small_number(n_thousands, True)} bin".strip())

    parts.append(spell_small_number(reminder))
    return " ".join(parts).strip()


def prompt_difficulty() -> Difficulty:
    """Prompt the difficulty from the user.

    Returns
    ----------
    difficulty : Difficulty
        A value from `Difficulty` enum.
    """
    return inquirer.select(
        message="Choose difficulty:",
        choices=[
            Choice(value=Difficulty.DIGITS, name="digits"),
            Choice(value=Difficulty.TENS, name="tens"),
            Choice(value=Difficulty.BASIC, name="basic"),
            Choice(value=Difficulty.ADVANCED, name="advanced"),
        ],
    ).execute()


def numbers(
    difficulty: Optional[Difficulty] = typer.Option(
        None, "--difficulty", help="Difficulty"
    )
) -> None:
    """Practice session for numbers.

    Parameters
    ----------
    difficulty : Optional[Difficulty]
        A value from `Difficulty` enum. Prompted from user if None.
    """
    if difficulty is None:
        difficulty = prompt_difficulty()

    number_generator: Callable
    match difficulty:
        case difficulty.DIGITS:
            number_generator = partial(random.choice, list(digits.keys()))
        case difficulty.TENS:
            number_generator = partial(random.choice, list(tens.keys()))
        case difficulty.BASIC:
            number_generator = partial(
                random.choice, list((digits | tens | more).keys())
            )
        case difficulty.ADVANCED:
            number_generator = partial(random.randrange, 10**12)

    prompter = PrompterInTheLanguage(Language.turkish)
    while True:
        number = number_generator()
        correct_answer = spell_number(number)
        print(
            f"Spell [yellow]{number:10_}[/yellow]. Press [blue]enter[/blue] to escape."
        )
        user_answer = prompter.prompt()
        if not user_answer:
            return
        if user_answer.split() == correct_answer.split():
            print("[green]Correct![/green]")
        else:
            print(
                f"[red]Incorrect![/red] Right answer:\n> [green]{correct_answer}[/green]"
            )


def main() -> None:
    """If open as a script, run make number function."""
    typer.run(numbers)


if __name__ == "__main__":
    main()



import typer

from practice_turkish.translation import translation
from practice_turkish.make_csv import make_dictionary as make_csv
from practice_turkish.number import numbers


def main() -> None:
    "Create Typer application and run it."
    app = typer.Typer()
    app.command(help="Practice translation")(translation)
    app.command(help="Create a new CSV dictionary")(make_csv)
    app.command(help="Practice numbers")(numbers)
    app()


if __name__ == "__main__":
    main()



from enum import Enum
from typing import Any

from InquirerPy import inquirer
from InquirerPy.base.control import Choice

from practice_turkish.languages.turkishinput import prompt_turkish
from practice_turkish.languages.russianinput import prompt_russian
from practice_turkish.languages.englishinput import prompt_english


class Language(str, Enum):
    "An enumeration used to represent natural languages."
    turkish = "turkish"
    russian = "russian"
    english = "english"


class PrompterInTheLanguage:
    """A class used to prompt from the user in specified language.

    Attributes
    ----------
    prompt_function : Callable[[str, str], str]
        One of the prompt_turkish, prompt_russian or prompt_english functions.

    Methods
    ----------
    prompt(self, message: str = "> ", additional_symbols: str = "") -> str:
        Prompt the user in the specified language
    """

    def __init__(self, language: Language) -> None:
        match language:
            case Language.turkish:
                self.prompt_function = prompt_turkish
            case Language.russian:
                self.prompt_function = prompt_russian
            case Language.english:
                self.prompt_function = prompt_english

    def prompt(
        self, message: str = "> ", additional_symbols: str = "", **kwargs: Any
    ) -> str:
        "Prompt the user in the specified language"
        return self.prompt_function(message, additional_symbols, **kwargs)


language_map = {
    Language.turkish: "Türkçe",
    Language.russian: "Русский",
    Language.english: "English",
}


def prompt_language(message: str) -> Language:
    """Prompt a language from the user by picking from values of `Language` enum.

    Parameters
    ----------
    message : str
        A text to be printed before the prompt.

    Returns
    ----------
    language : Language
        A value of `Language` enum.
    """
    return inquirer.select(
        message=message,
        choices=[Choice(value=key, name=value) for key, value in language_map.items()],
    ).execute()


def prompt_way_of_translation(language_a: Language, language_b: Language) -> bool:
    """Prompt a way of translation: language A ➔ language B or vice versa.

    Parameters
    ----------
    language_a : Language
        Language A.
    language_b : Language
        Language B

    Returns
    ----------
    a2b : bool
        True, if the user picked to translate from language A to language B.
        False, if the user picked to translate from language B to language A.
    """
    la, lb = language_map[language_a], language_map[language_b]
    max_len = max(len(la), len(lb))
    la, lb = la.ljust(max_len), lb.ljust(max_len)
    return inquirer.select(
        message="Choose language you want to translate to.",
        choices=[
            Choice(value=True, name=f"{la} → {lb}"),
            Choice(value=False, name=f"{lb} → {la}"),
        ],
    ).execute()
